//
// Gb.cpp
//
// Encode/decode for GB2312.1980 
//
// Code adapted from the GNU ICONV Library.
// 

/* Copyright (C) 1999-2000 Free Software Foundation, Inc.

   The GNU ICONV Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU ICONV Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU ICONV Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "ufcodec.h"

#include <e32def.h>
#include <e32uid.h>

typedef struct {
  unsigned short indx; /* index into big table */
  unsigned short used; /* bitmask of used entries */
} Summary16;

#include "gb2312.h"

// ------------------------------------------------------------

class Uni {
public:
  enum { KIllegal = 0xffff, KIncomplete = 0xf800, KHeadless = 0xf801 };
  static TText16 WChar(const TDesC8 &aDes, TInt &aPos);
  static void AppendUTF8(TDes8 &aUtf, TText16 aWChar);
};

LOCAL_D const TUint8 bytesFromUTF8[256] = {
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
  2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5};

LOCAL_D const TText16 firstByteMark[7] = {
  0x00, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0 };

// convert UTF8 to Unicode
// special cases:  end of string in UTF8 sequence --> KIncomplete
//                 UTF8 sequence stops prematurely --> KIllegal
//                 UTF8 sequence starts with trailer --> KHeadless
TText16 Uni::WChar(const TDesC8 &aDes, TInt &aPos)
{
  TUint wch = aDes[aPos++];
  if ((wch & 0xc0) == 0x80) {
    // KHeadless
    while (aPos < aDes.Length() && (aDes[aPos] & 0xc0) == 0x80)
      aPos++;
    return KHeadless;
  }
  TInt extraBytes = bytesFromUTF8[wch & 0xff];
  wch -= firstByteMark[extraBytes];
  while (extraBytes--) {
    if (aPos >= aDes.Length())
      return TText16(KIncomplete);
    if ((aDes[aPos] & 0xc0) != 0x80)
      // return TText16(aDes[aPos - 1]);
      return TText16(KIllegal);
    wch <<= 6;
    wch |= aDes[aPos++] & 0x3f; 
  }
  if (wch == 0x7ff) return 0x6708; // moon ideograph
  return TText16(wch);
}

// convert Unicode to UTF8
void Uni::AppendUTF8(TDes8 &aUtf, TText16 wch)
{
  if (wch < 0x80) {
    aUtf.Append(TChar(wch));
  } else if (wch < 0x800) {
    aUtf.Append(TChar(((wch & 0x7c0) >> 6) | 0xc0));
    aUtf.Append(TChar((wch & 0x03f) | 0x80));
  } else {
    // We never need to write UCS larger than 0x10000,
    // so now it's three bytes 
    aUtf.Append(TChar(((wch & 0x0f000) >> 12) | 0xe0));
    aUtf.Append(TChar(((wch & 0xfc0) >> 6) | 0x80));
    aUtf.Append(TChar((wch & 0x03f) | 0x80));
  }
}

// ------------------------------------------------------------

class CGbEncoder : public CUniCodec {
public:
  virtual TText16 DecodeWChar(TText16 aChar);
  virtual TText16 EncodeWChar(TText16 aChar);
  virtual HBufC8 *EncodeTextL(const TDesC8 &aText, TBool aSkipUnknown);
  virtual HBufC8 *DecodeTextL(const TDesC8 &aText);

  virtual TUid Uid() const { return TUid::Uid(0x10001860); }
  virtual TInt ApiLevel() const { return 17; }
  virtual TTime Version() const
  { return TTime(TDateTime(2000, ENovember, 1, 0, 0, 0, 0)); }
  virtual TInt MaximumSizeInBytes() const { return 2; }
  virtual TInt MaximumSizeOfUnicodeRepresentation() const { return 1; }
  virtual void EncodingContextReset() { }
  virtual TBool EncodingContextNonTrivial() const { return EFalse; }
  virtual TInt EncodingContextData() const { return 0; }
  virtual void DecodingContextReset() { iChar = 0; } 
  virtual TBool DecodingContextNonTrivial() const { return (iChar != 0); }
  virtual TInt DecodingContextData() const 
    { return (iChar != 0) ? 1 : 0; }
  virtual void EncodeL(MUniCodecUnicodeSource *source,
		       MUniCodecNativeSink &sink);
  virtual TInt DecodeL(MUniCodecNativeSource *source,
                       MUniCodecUnicodeSink &sink);
private:
  TInt iChar;
};

// ------------------------------------------------------------

TText16 CGbEncoder::EncodeWChar(TText16 wc)
{
  if (wc < 0x80) return wc;
  const Summary16 *summary = NULL;
  int i;
  if (wc >= 0x0000 && wc < 0x0460)
    summary = &gb2312_uni2indx_page00[i = (wc>>4)];
  else if (wc >= 0x2000 && wc < 0x2650)
    summary = &gb2312_uni2indx_page20[i = (wc>>4)-0x200];
  else if (wc >= 0x3000 && wc < 0x3230)
    summary = &gb2312_uni2indx_page30[i = (wc>>4)-0x300];
  else if (wc >= 0x4e00 && wc < 0x9cf0)
    summary = &gb2312_uni2indx_page4e[i = (wc>>4)-0x4e0];
  else if (wc >= 0x9e00 && wc < 0x9fb0)
    summary = &gb2312_uni2indx_page9e[i = (wc>>4)-0x9e0];
  else if (wc >= 0xff00 && wc < 0xfff0)
    summary = &gb2312_uni2indx_pageff[i = (wc>>4)-0xff0];
  if (summary) {
    unsigned short used = summary->used;
    unsigned int i = wc & 0x0f;
    if (used & ((unsigned short) 1 << i)) {
      unsigned short c;
      /* Keep in `used' only the bits 0..i-1. */
      used &= ((unsigned short) 1 << i) - 1;
      /* Add `summary->indx' and the number of bits set in `used'. */
      used = (used & 0x5555) + ((used & 0xaaaa) >> 1);
      used = (used & 0x3333) + ((used & 0xcccc) >> 2);
      used = (used & 0x0f0f) + ((used & 0xf0f0) >> 4);
      used = (used & 0x00ff) + (used >> 8);
      c = gb2312_2charset[summary->indx + used];
      return TText16(c | 0x8080);
    }
  }
  return 0;
}

TText16 CGbEncoder::DecodeWChar(TText16 aGb)
{
  if (aGb < 0x80) return aGb;
  if ((aGb & 0x8080) != 0x8080) return 0;
  aGb &= 0x7f7f;
  unsigned char c1 = aGb >> 8;
  unsigned char c2 = aGb & 0xff;
  if ((c1 >= 0x21 && c1 <= 0x29) || (c1 >= 0x30 && c1 <= 0x77)) {
    if (c2 >= 0x21 && c2 < 0x7f) {
      unsigned int i = 94 * (c1 - 0x21) + (c2 - 0x21);
      unsigned short wc = 0xfffd;
      if (i < 1410) {
	if (i < 831)
	  wc = gb2312_2uni_page21[i];
      } else {
	if (i < 8178) {
	  i -= 1410;
	  wc = gb2312_2uni_page30[i];
	}
      }
      return TText16(wc);
    }
  }
  return 0xfffd;
}

// --------------------------------------------------------------------

HBufC8* CGbEncoder::EncodeTextL(const TDesC8 &aText, TBool aSkipUnknown)
{
  HBufC8 *anEncoded = HBufC8::NewL(6 * aText.Length());
  TPtr8 aBuf = anEncoded->Des();
  TInt i = 0;
  while (i < aText.Length()) {
    TText16 aUni = Uni::WChar(aText, i);
    TText16 aEnc = EncodeWChar(aUni);
    if (aEnc) {
      if (aEnc >= 0x100) {
	aBuf.Append(TChar(aEnc >> 8));
      }
      aBuf.Append(TChar(aEnc & 0xff));
    } else {
      if (!aSkipUnknown)
	aBuf.AppendFormat(_L("\\u%04x"), aUni);
    }
  }
  return anEncoded;
}

HBufC8* CGbEncoder::DecodeTextL(const TDesC8 &aText)
{
  HBufC8 *aDecoded = HBufC8::NewL(3 * aText.Length());
  TPtr8 aBuf = aDecoded->Des();
  TInt j = 0;
  while (j < aText.Length()) {
    TUint aGb = aText[j++];
    if (aGb < 0x80 || j == aText.Length()) {
      // pass
    } else {
      aGb <<= 8;
      aGb |= aText[j++];
    }
    Uni::AppendUTF8(aBuf, DecodeWChar(TText16(aGb)));
  }
  return aDecoded;
}

// ---------------------------------------------------------------------

void CGbEncoder::EncodeL(MUniCodecUnicodeSource *source, 
			 MUniCodecNativeSink &sink)
{
  if (!source)
    // no context to flush
    return;

  for (;;) {
    TInt wch = source->GetL();
    if (wch < 0) 
      return;
    if (!wch) {
      sink.PutL(wch);
    } else {
      TText16 aEnc = EncodeWChar(wch);
      if (aEnc) {
	if (aEnc >= 0x100)
	  sink.PutL(aEnc >> 8);
	if (sink.PutL(aEnc & 0xff))
	  return;
      } else {
	if (sink.LossyL(wch))
	  return;
      }
    }
  }
}

TInt CGbEncoder::DecodeL(MUniCodecNativeSource *source, 
			 MUniCodecUnicodeSink &sink)
{
  if (!source) {
    if (iChar != 0)
      sink.IllegalL(iChar);
    return 0;
  }
  
  for (;;) {
    if (!iChar) {
      TInt ch = source->GetL(ETrue);
      if (ch < 0)
	return 0;
      if (ch < 0x80) {
	if (sink.PutL(DecodeWChar(TText16(ch))))
	  return 0;
      } else {
	iChar = ch << 8;
      }
    } else {
      // iChar != 0
      TInt ch = source->GetL(EFalse);
      if (ch < 0)
	return 1;
      ch |= iChar;
      iChar = 0;
      if (sink.PutL(DecodeWChar(TText16(ch))))
	return 0;
    }
  }
}

// --------------------------------------------------------------------

// The E32Dll() entry point function

GLDEF_C TInt E32Dll(TDllReason /*aReason*/)
{
  return(KErrNone);
}

// Function to construct a CUniCodec object. Note that this function
// is exported at ordinal 1.

EXPORT_C CUniEncoder* NewEncoder()
{
  return new CGbEncoder;
}
