//
// Ufcjis.cpp
//
// Otfried Cheong
//
// Library for JIS - based encodings
// 

#include "ufcodec.h"
#include "ufcjis.h"

#include <e32std.h>

#include <e32def.h>
#include <e32uid.h>

GLREF_D "C" const TUint16 sjis_map[];
GLREF_D "C" const TUint16 * const uni2sjis[0x100];

// --------------------------------------------------------------------

class Uni {
public:
  enum { KIllegal = 0xffff, KIncomplete = 0xf800, KHeadless = 0xf801 };
  static TText16 WChar(const TDesC8 &aDes, TInt &aPos);
  static void AppendUTF8(TDes8 &aUtf, TText16 aWChar);
  static TText16 SjisToUni(TText16 wch);
  static TText16 UniToSjis(TText16 wch);
  static TText16 UniToJis(TText16 aUni);
  static TText16 JisToUni(TText16 aJis);
};

static const TText16 KWeekdays[] = { 
  0x6708, 0x706b, 0x6c34, 0x6728, 0x91d1, 0x571f, 0x65e5, 0 };

GLDEF_C const TUint8 bytesFromUTF8[256] = {
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
  2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5};

LOCAL_C const TText16 firstByteMark[7] = {
  0x00, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0 };

// convert UTF8 to Unicode
// special cases:  end of string in UTF8 sequence --> KIncomplete
//                 UTF8 sequence stops prematurely --> KIllegal
//                 UTF8 sequence starts with trailer --> KHeadless
TText16 Uni::WChar(const TDesC8 &aDes, TInt &aPos)
{
  TUint wch = aDes[aPos++];
  if (wch >= 0xf8) return KWeekdays[wch - 0xf8];
  if ((wch & 0xc0) == 0x80) {
    // KHeadless
    while (aPos < aDes.Length() && (aDes[aPos] & 0xc0) == 0x80)
      aPos++;
    return KHeadless;
  }
  TInt extraBytes = bytesFromUTF8[wch & 0xff];
  wch -= firstByteMark[extraBytes];
  while (extraBytes--) {
    if (aPos >= aDes.Length())
      return TText16(KIncomplete);
    if ((aDes[aPos] & 0xc0) != 0x80)
      // return TText16(aDes[aPos - 1]);
      return TText16(KIllegal);
    wch <<= 6;
    wch |= aDes[aPos++] & 0x3f; 
  }
  if (wch == 0x7ff) return 0x6708; // moon ideograph
  return TText16(wch);
}

// convert Unicode to UTF8
void Uni::AppendUTF8(TDes8 &aUtf, TText16 wch)
{
  if (wch < 0x80) {
    aUtf.Append(TChar(wch));
  } else if (wch < 0x800) {
    aUtf.Append(TChar(((wch & 0x7c0) >> 6) | 0xc0));
    aUtf.Append(TChar((wch & 0x03f) | 0x80));
  } else {
    // We never need to write UCS larger than 0x10000,
    // so now it's three bytes 
    aUtf.Append(TChar(((wch & 0x0f000) >> 12) | 0xe0));
    aUtf.Append(TChar(((wch & 0xfc0) >> 6) | 0x80));
    aUtf.Append(TChar((wch & 0x03f) | 0x80));
  }
}

// ------------------------------------------------------------

TText16 Uni::UniToSjis(TText16 aUni)
{
  TUint aPage = aUni >> 8;
  const TUint16 *p = uni2sjis[aPage];
  return TText16( p ? p[aUni & 0xff] : 0 );
}

TText16 Uni::UniToJis(TText16 aUni)
{
  TUint sjis = UniToSjis(aUni);
  if (!sjis || sjis < 0x80)
    return TText16(sjis);
  if (sjis < 0x100) 
    // kankaku kana do not exist in JIS
    return 0;  
  TUint c1 = sjis >> 8;
  TUint c2 = sjis & 0xff;
  TInt adjust = c2 < 159;
  TInt rowOffset = c1 < 160 ? 112 : 176;
  TInt cellOffset = adjust ? (c2 > 127 ? 32 : 31) : 126;
  TUint jis = (((c1 - rowOffset) << 1) - adjust) << 8;
  jis += (c2 - cellOffset);
  return TText16(jis);
}

LOCAL_D const TUint16 hara_map[] = {
  0x0000, 0x3002, 0x300c, 0x300d, 0x3001, 0x30fb, 0x30f2, 0x30a1,
    0x30a3, 0x30a5, 0x30a7, 0x30a9, 0x30e3, 0x30e5, 0x30e7, 0x30c3,
    0x30fc, 0x30a2, 0x30a4, 0x30a6, 0x30a8, 0x30aa, 0x30ab, 0x30ad,
    0x30af, 0x30b1, 0x30b3, 0x30b5, 0x30b7, 0x30b9, 0x30bb, 0x30bd,
    0x30bf, 0x30c1, 0x30c4, 0x30c6, 0x30c8, 0x30ca, 0x30cb, 0x30cc,
    0x30cd, 0x30ce, 0x30cf, 0x30d2, 0x30d5, 0x30d8, 0x30db, 0x30de,
    0x30df, 0x30e0, 0x30e1, 0x30e2, 0x30e4, 0x30e6, 0x30e8, 0x30e9,
    0x30ea, 0x30eb, 0x30ec, 0x30ed, 0x30ef, 0x30f0, 0x30f1, 0x30f2,
    0x30f3, 0x309b, 0x309c
    };

TText16 Uni::SjisToUni(TText16 aSjis)
{
  if (aSjis <= 0x80 || aSjis == 0xa0 ||
      (0xfd <= aSjis && aSjis < 0x100)) {
    switch (aSjis) {
    case 0x80:
      aSjis = 0x5c;
      break;
    case 0xfd:
      aSjis = 0xa9;
      break;
    case 0xfe:
      aSjis = 0x2122;
      break;
    case 0xff:
      aSjis = 0x2026;
      break;
    default:
      break;
    }
    return aSjis;
  } else if (0xa1 <= aSjis && aSjis < 0xe0) {
    return hara_map[aSjis - 0xa0];
  } else {
    // two-byte SJIS
    TUint16 aTwo = TUint16(aSjis & 0xff);
    aSjis >>= 8;
    if (aTwo > 0x7f)
      aTwo--;
    aTwo -= 0x40;
    if (aSjis >= 0xe0)
      aSjis -= (0xe0 - 0xa0);
    aSjis -= 0x81;
    aSjis *= (0xfd - 0x40) - 1;
    aSjis += aTwo;
    // all SJIS should be convertable!
    if (aSjis < 11280)
      return sjis_map[aSjis];
    else
      return 0xfffd;
  }
}

TText16 Uni::JisToUni(TText16 aJis)
{
  if (aJis < 0x100)
    return aJis;
  TUint c1 = aJis >> 8;
  TUint c2 = aJis & 0xff;
  TInt rowOffset = c1 < 95 ? 112 : 176;
  TInt cellOffset = c1 % 2 ? (c2 > 95 ? 32 : 31) : 126;
  TUint aSjis = (((c1 + 1) >> 1) + rowOffset) << 8;
  aSjis += (c2 + cellOffset);
  return SjisToUni(TText16(aSjis));
}

// ------------------------------------------------------------

class CJisCodec : public CUniCodec {
public:
  virtual HBufC8 *EncodeTextL(const TDesC8 &aText, TBool aSkipUnknown);
  virtual HBufC8 *DecodeTextL(const TDesC8 &aText);
  virtual TInt ApiLevel() const { return 17; }
};

// ------------------------------------------------------------

class TUnicodeSource : public MUniCodecUnicodeSource {
public:
  TUnicodeSource(const TPtrC &aPtr) : iPtr(aPtr), iIndex(0) { }
  TInt GetL();
private:
  const TPtrC &iPtr;
  TInt iIndex;
};

TInt TUnicodeSource::GetL()
{
  if (iIndex < iPtr.Length())
     return Uni::WChar(iPtr, iIndex);
  else
     return -1;
}

class TNativeSink : public MUniCodecNativeSink {
public:
  TNativeSink(TPtr &aPtr, TBool aSkip) : iPtr(aPtr), iSkip(aSkip) { } 
  TBool PutL(TInt byte);
  TBool LossyL(TUniChar wc);
private:
  TPtr &iPtr;
  TBool iSkip;
};

TBool TNativeSink::PutL(TInt byte)
{
  iPtr.Append(TChar(byte));
  return EFalse;
}

TBool TNativeSink::LossyL(TUniChar wc)
{
  if (!iSkip) {
    iPtr.AppendFormat(_L("\\u%04x"), wc);
  }
  return EFalse;
}

// --------------------------------------------------------------------

HBufC8 *CJisCodec::EncodeTextL(const TDesC8 &aText, TBool aSkipUnknown)
{
  HBufC8 *hbuf = HBufC8::NewLC(aText.Length() * Max(6, MaximumSizeInBytes()));
  TUnicodeSource source(aText);
  TPtr8 sinkbuf(hbuf->Des());
  TNativeSink sink(sinkbuf, aSkipUnknown);
  EncodingContextReset();
  EncodeL(&source, sink);
  EncodeL(0, sink);
  CleanupStack::Pop(); // hbuf;
  return hbuf;
}

// --------------------------------------------------------------------

class TNativeSource : public MUniCodecNativeSource {
public:
  TNativeSource(const TPtrC &aPtr) : iPtr(aPtr), iIndex(0) { }
  virtual TInt GetL(TBool first);
private:
  const TPtrC &iPtr;
  TInt iIndex;
};

TInt TNativeSource::GetL(TBool /* first */)
{
  if (iIndex < iPtr.Length())
    return iPtr[iIndex++];
  else
    return -1;
}

class TUtfSink : public MUniCodecUnicodeSink {
public:
  TUtfSink(TPtr &aPtr) : iPtr(aPtr) { } 
  TBool PutL(TUniChar wc);
  TBool IllegalL(TUint32 bytes);
private:
  TPtr &iPtr;
};

TBool TUtfSink::PutL(TUniChar wch) 
{
  Uni::AppendUTF8(iPtr, wch);
  return EFalse;
}

// ------------------------------------------------------------


TBool TUtfSink::IllegalL(TUint32 /* bytes */)
{
  Uni::AppendUTF8(iPtr, 0xfffd);
  return EFalse;
}

// --------------------------------------------------------------------

HBufC8 *CJisCodec::DecodeTextL(const TDesC8 &aText)
{
  HBufC8 *hbuf = HBufC8::NewLC(aText.Length() * 6);
  TNativeSource source(aText);
  TPtr8 sinkbuf(hbuf->Des());
  TUtfSink sink(sinkbuf);
  DecodingContextReset();
  DecodeL(&source, sink);
  DecodeL(0, sink);
  CleanupStack::Pop(); // hbuf;
  return hbuf;
}

// --------------------------------------------------------------------

class CEucJpEncoder : public CJisCodec {
public:
  virtual TText16 DecodeWChar(TText16 wc);
  virtual TText16 EncodeWChar(TText16 wc);
  virtual TUid Uid() const { return TUid::Uid(0x10000e29); }
  virtual TTime Version() const
  { return TTime(TDateTime(2000, ENovember, 3, 0, 0, 0, 0)); }
  virtual TInt MaximumSizeInBytes() const { return 2; }
  virtual TInt MaximumSizeOfUnicodeRepresentation() const { return 6; }
  virtual void EncodingContextReset() { }
  virtual TBool EncodingContextNonTrivial() const { return EFalse; }
  virtual TInt EncodingContextData() const { return 0; }
  virtual void DecodingContextReset() { iChar = 0; } 
  virtual TBool DecodingContextNonTrivial() const { return (iChar != 0); }
  virtual TInt DecodingContextData() const 
    { return (iChar != 0) ? 1 : 0; }
  virtual void EncodeL(MUniCodecUnicodeSource *source,
		       MUniCodecNativeSink &sink);
  virtual TInt DecodeL(MUniCodecNativeSource *source,
                       MUniCodecUnicodeSink &sink);
private:
  TInt iChar;
};

// ------------------------------------------------------------

TText16 CEucJpEncoder::EncodeWChar(TText16 aUni)
{
  TText16 jis = Uni::UniToJis(aUni);
  if (jis < 0x100)
    return jis;
  return TText16(jis + 0x8080);
}

TText16 CEucJpEncoder::DecodeWChar(TText16 aEuc)
{
  if (aEuc < 0x100) 
    return aEuc;
  return Uni::JisToUni(TText16(aEuc - 0x8080));
}

void CEucJpEncoder::EncodeL(MUniCodecUnicodeSource *source, 
			    MUniCodecNativeSink &sink)
{
  if (!source)
    // no context to flush
    return;

  for (;;) {
    TInt wch = source->GetL();
    if (wch < 0) 
      return;
    TText16 aEnc = EncodeWChar(wch);
    if (aEnc) {
      if (aEnc >= 0x100)
	sink.PutL(aEnc >> 8);
      if (sink.PutL(aEnc & 0xff))
	return;
    } else {
      if (sink.LossyL(wch))
	return;
    }
  }
}

TInt CEucJpEncoder::DecodeL(MUniCodecNativeSource *source, 
			    MUniCodecUnicodeSink &sink)
{
  if (!source) {
    if (iChar != 0)
      sink.IllegalL(iChar);
    return 0;
  }
  
  for (;;) {
    if (!iChar) {
      TInt ch = source->GetL(ETrue);
      if (ch < 0)
	return 0;
      if (ch < 0x80) {
	if (sink.PutL(DecodeWChar(ch)))
	  return 0;
      } else {
	iChar = ch << 8;
      }
    } else {
      // iChar != 0
      TInt ch = source->GetL(EFalse);
      if (ch < 0)
	return 1;
      ch |= iChar;
      iChar = 0;
      if (sink.PutL(DecodeWChar(ch)))
	return 0;
    }
  }
}

EXPORT_C CUniCodec* NewEucJpEncoder()
{
  return new CEucJpEncoder;
}

// --------------------------------------------------------------------

class CShiftJisEncoder : public CJisCodec {
public:
  virtual TText16 DecodeWChar(TText16 wc) { return Uni::SjisToUni(wc); }
  virtual TText16 EncodeWChar(TText16 wc) { return Uni::UniToSjis(wc); }
  virtual TUid Uid() const { return TUid::Uid(0x10000e26); }
  virtual TTime Version() const
  { return TTime(TDateTime(2000, ENovember, 3, 0, 0, 0, 0)); }
  virtual TInt MaximumSizeInBytes() const { return 2; }
  virtual TInt MaximumSizeOfUnicodeRepresentation() const { return 6; }
  virtual void EncodingContextReset() { }
  virtual TBool EncodingContextNonTrivial() const { return EFalse; }
  virtual TInt EncodingContextData() const { return 0; }
  virtual void DecodingContextReset() { iChar = 0; } 
  virtual TBool DecodingContextNonTrivial() const { return (iChar != 0); }
  virtual TInt DecodingContextData() const 
    { return (iChar != 0) ? 1 : 0; }
  virtual void EncodeL(MUniCodecUnicodeSource *source,
		       MUniCodecNativeSink &sink);
  virtual TInt DecodeL(MUniCodecNativeSource *source,
                       MUniCodecUnicodeSink &sink);
private:
  TInt iChar;
};

// ------------------------------------------------------------

void CShiftJisEncoder::EncodeL(MUniCodecUnicodeSource *source, 
			       MUniCodecNativeSink &sink)
{
  if (!source)
    // no context to flush
    return;

  for (;;) {
    TInt wch = source->GetL();
    if (wch < 0) 
      return;
    TText16 aEnc = Uni::UniToSjis(wch);
    if (aEnc) {
      if (aEnc >= 0x100)
	sink.PutL(aEnc >> 8);
      if (sink.PutL(aEnc & 0xff))
	return;
    } else {
      if (sink.LossyL(wch))
	return;
    }
  }
}

TInt CShiftJisEncoder::DecodeL(MUniCodecNativeSource *source, 
			       MUniCodecUnicodeSink &sink)
{
  if (!source) {
    if (iChar != 0)
      sink.IllegalL(iChar);
    return 0;
  }
  
  for (;;) {
    if (!iChar) {
      TInt ch = source->GetL(ETrue);
      if (ch < 0)
	return 0;
      if (ch <= 0x80 || (0xa1 <= ch && ch < 0xe0) || 0xfd <= ch) {
	if (sink.PutL(Uni::SjisToUni(ch)))
	  return 0;
      } else {
	iChar = ch << 8;
      }
    } else {
      // iChar != 0
      TInt ch = source->GetL(EFalse);
      if (ch < 0)
	return 1;
      ch |= iChar;
      iChar = 0;
      if (sink.PutL(Uni::SjisToUni(ch)))
	return 0;
    }
  }
}

EXPORT_C CUniCodec* NewShiftJisEncoder()
{
  return new CShiftJisEncoder;
}

// --------------------------------------------------------------------

class CIsoJpEncoder : public CJisCodec {
public:
  virtual TText16 DecodeWChar(TText16 wc) { return Uni::JisToUni(wc); }
  virtual TText16 EncodeWChar(TText16 wc) { return Uni::UniToJis(wc); }
  virtual TUid Uid() const { return TUid::Uid(0x10000e28); }
  virtual TTime Version() const
  { return TTime(TDateTime(2000, ENovember, 3, 0, 0, 0, 0)); }
  virtual TInt MaximumSizeInBytes() const { return 7; }
  virtual TInt MaximumSizeOfUnicodeRepresentation() const { return 1; }
  virtual void EncodingContextReset()
  { iEncodingContextKanji = EFalse; }
  virtual TBool EncodingContextNonTrivial() const
  { return iEncodingContextKanji; }
  virtual TInt EncodingContextData() const { return 0; }
  virtual void DecodingContextReset()
  { iChar = 0; iPending = 0; 
  iDecodingContextKanji = iDecodingContextKana = EFalse; } 
  virtual TBool DecodingContextNonTrivial() const 
  { return (iPending != 0) || (iChar != 0) || 
      iDecodingContextKanji || iDecodingContextKana; }
  virtual TInt DecodingContextData() const 
  { return iPending + (iChar != 0 ? 1 : 0); }
  virtual void EncodeL(MUniCodecUnicodeSource *source,
		       MUniCodecNativeSink &sink);
  virtual TInt DecodeL(MUniCodecNativeSource *source,
                       MUniCodecUnicodeSink &sink);
private:
  void FlushPartialSequence(MUniCodecUnicodeSink &sink);
private:
  TBool iEncodingContextKanji;
  TBool iDecodingContextKanji;
  TBool iDecodingContextKana;
  TInt iPending;
  TText8 iPendingChar[4];
  TInt iChar;
};

// ------------------------------------------------------------

void CIsoJpEncoder::EncodeL(MUniCodecUnicodeSource *source, 
			    MUniCodecNativeSink &sink)
{
  if (!source) {
    if (iEncodingContextKanji) {
      // single byte mode
      sink.PutL(0x1b);
      sink.PutL('(');
      sink.PutL('B');
    }
    return;
  }

  for (;;) {
    TInt wch = source->GetL();
    if (wch < 0) 
      return;
    TText16 aEnc = Uni::UniToJis(wch);
    if (!aEnc) {
	if (iEncodingContextKanji) {
	  iEncodingContextKanji = EFalse;
	  // single byte mode
	  sink.PutL(0x1b);
	  sink.PutL('(');
	  sink.PutL('B');
	}
	if (sink.LossyL(wch))
	  return;
    } else {
      if (aEnc < 0x100) {
	if (iEncodingContextKanji) {
	  iEncodingContextKanji = EFalse;
	  // single byte mode
	  sink.PutL(0x1b);
	  sink.PutL('(');
	  sink.PutL('B');
	}
	if (sink.PutL(aEnc))
	  return;
      } else {
	if (!iEncodingContextKanji) {
	  iEncodingContextKanji = ETrue;
	  // double byte mode
	  sink.PutL(0x1b);
	  sink.PutL('$');
	  sink.PutL('B');
	}
	sink.PutL(aEnc >> 8);
	if (sink.PutL(aEnc & 0xff))
	  return;
      }
    }
  }
}

void CIsoJpEncoder::FlushPartialSequence(MUniCodecUnicodeSink &sink)
{
  // flush out partial control sequence
  for (TInt i = 0; i < iPending; )
    sink.PutL(iPendingChar[i++]);
  iPending = 0;
}

// sequences recognized on decoding:
// ESC $ B and ESC $ @   -> Kanji, not Kana
// ESC ( B and ESC ( J   -> not Kanji, not Kana
// ESC ( I               -> Kana (outrageous: Netscape, Outlook Express)
// \n, \r                -> not Kanji, not Kana
// 0x0e                  -> Kana (outrageous: MS internet mail)
// 0x0f                  -> not Kana

// decoding:

TInt CIsoJpEncoder::DecodeL(MUniCodecNativeSource *source, 
			    MUniCodecUnicodeSink &sink)
{
  if (!source) {
    if (iPending)
      FlushPartialSequence(sink);
    if (iChar)
      sink.IllegalL(iChar);
    return 0;
  }
  
  for (;;) {
    if (!iPending && !iChar) {
      TInt ch = source->GetL(ETrue);
      if (ch < 0)
	return 0;
      if (ch == 0x0e) {
	iDecodingContextKana = ETrue;
      } else if (ch == 0x0f) {
	iDecodingContextKana = EFalse;
      } else if (ch == 0x1b) {
	// wait for ESC ... 
	iPending = 1;
      } else if (ch == '\r' || ch == '\n') {
	iDecodingContextKanji = EFalse;
	iDecodingContextKana = EFalse;
	if (sink.PutL(ch))
	  return 0;
      } else if (iDecodingContextKana) {
	// outrageous kana character
	if (0x21 <= ch && ch <= 0x5f) {
	  // ignore if it doesn't belong to 7bits kana.
	  if (sink.PutL(TText16(ch + 0x80 - 0x40 + 0xff00)))
	    return 0;
	}
      } else if (!iDecodingContextKanji) {
	if (sink.PutL(ch))
	  return 0;
      } else if (iDecodingContextKanji) {
	// interpret two bytes
	iChar = ch << 8;
      }
    } else if (iPending) {
      // in ESC ... sequence
      TInt ch = source->GetL(EFalse);
      if (ch < 0)
	return 0; // not in middle of character!
      if ((iPending == 1 && ch == '$') ||
	  (iPending == 1 && ch == '(') ||
	  (iPending == 2 && ch == 'B') ||
	  (iPending == 2 && iPendingChar[1] == '$' && ch == '@') ||
	  (iPending == 2 && iPendingChar[1] == '(' && ch == 'J') ||
	  (iPending == 2 && iPendingChar[1] == '(' && ch == 'I')) {
	iPending++;
      } else {
	// different sequence
	FlushPartialSequence(sink);
	if (sink.PutL(ch))
	  return 0;
      }
      if (iPending == 3) {
	// sequence complete, process and discard it
	if (iPendingChar[2] == 'I') {
	  iDecodingContextKana = ETrue;
	} else {
	  iDecodingContextKanji = (iPendingChar[1] == '$');
	  iDecodingContextKana = EFalse;
	}
	iPending = 0;
      }
    } else {
      // waiting for second byte
      TInt ch = source->GetL(EFalse);
      if (ch < 0)
	return 1;
      ch |= iChar;
      iChar = 0;
      if (sink.PutL(DecodeWChar(ch & 0x7f7f)))
	return 0;
    }
  }
}

EXPORT_C CUniCodec* NewIsoJpEncoder()
{
  return new CIsoJpEncoder;
}

// ---------------------------------------------------------------------

// The E32Dll() entry point function

GLDEF_C TInt E32Dll(TDllReason /*aReason*/)
{
  return(KErrNone);
}

/* 
HBufC8* CJisEncoder::DecodeText(const TDesC8 &aText)
{
  HBufC8 *aDecoded = HBufC8::NewL((3 * aText.Length() + 2)/2);
  TPtr8 aBuf = aDecoded->Des();
  TInt j = 0;
  TBool aKana = EFalse;
  TBool aKanji = EFalse;
  while (j < aText.Length()) {
    TUint aJis = aText[j++];
    if (aJis == 0x1b) {
      aJis = aText[j++];
      if (aJis == '$') {
	aJis = aText[j++];
	if (aJis == 'B' || aJis == '@') {
	  aKanji = ETrue;
	  aKana = EFalse;
	} else {
	  aBuf.Append(TChar(0x1b));
	  aBuf.Append(TChar('$'));
	  aBuf.Append(TChar(aJis));
	}
      } else if (aJis == '(') {
	aJis = aText[j++];
	if (aJis == 'B' || aJis == 'J') {
	  aKanji = EFalse;
	  aKana = EFalse;
	} else if (aJis == 'I') {
	  // ESC-(-I is outrageous kana-in used by
	  // Netscape and Outlook Express.
	  aKana = ETrue;
	} else {
	  aBuf.Append(TChar(0x1b));
	  aBuf.Append(TChar('('));
	  aBuf.Append(TChar(aJis));
	}
      } else {
	  aBuf.Append(TChar(0x1b));
	  aBuf.Append(TChar(aJis));
      }
    } else if (aJis == '\n' || aJis == '\r') {
      aKanji = aKana = EFalse;
      aBuf.Append(TChar(aJis));
    } else if (aJis == 0x0e) {
      // Outrageous Kana-in used by MS internet mail
      aKana = ETrue;
    } else if (aJis == 0x0f) {
      // outrageous Kana-out used by MS internet mail
      aKana = EFalse;
    } else if (aKana) {
      // outrageous kana character
      if (0x21 <= aJis && aJis <= 0x5f) {
	// ignore if it doesn't belong to 7bits kana.
	Uni::AppendUTF8(aBuf, TText16(aJis + 0x80 - 0x40 + 0xff00));
      }
    } else if (!aKanji) {
      aBuf.Append(TChar(aJis));
    } else {
      // Kanji
      aJis <<= 8;
      aJis += aText[j++];
      Uni::AppendUTF8(aBuf, DecodeWChar(TText16(aJis)));
    }
  }
  return aDecoded;
}

*/
